perl -I../lib 00-error.t 1
perl -I../lib 00-obj.t 1
perl -I../lib 01-run.t 1
perl -I../lib 01-run_echo.t 1
perl -I../lib 01-run_output=merged.t 1
perl -I../lib 01-run_output=quiet.t 1
perl -I../lib 01-run_output=stderr.t 1
perl -I../lib 01-run_output=stdout.t 1
perl -I../lib 01-script.t 1
perl -I../lib 01-script=run.t 1
perl -I../lib 01-script=run_output=stdout.t 1
perl -I../lib 01-script=script.t 1
perl -I../lib 01-script=simple.t 1
perl -I../lib 02-run.t 1
perl -I../lib 02-run_continue.t 1
perl -I../lib 02-run_display.t 1
perl -I../lib 02-script=run.t 1
perl -I../lib 03-run.t 1
perl -I../lib 03-script=run.t 1
perl -I../lib 04-run.t 1
perl -I../lib 04-script=run.t 1
perl -I../lib 05-run.t 1
perl -I../lib 05-run_continue.t 1
perl -I../lib 06-run.t 1
perl -I../lib 06-script=run.t 1
perl -I../lib 07-run.t 1
perl -I../lib 07-script=run.t 1
perl -I../lib 08-run.t 1
perl -I../lib 08-script=run.t 1
perl -I../lib 09-run.t 1
perl -I../lib 09-script=run.2.t 1
perl -I../lib 09-script=run.t 1
perl -I../lib 99-cleanup.t 1
